# Copyright [2018-2020] Peter Krenesky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from stringcolor import cs


DEFAULT_COLORS = {
    "DEBUG": "white",
    "INFO": "white",
    "WARNING": "yellow",
    "ERROR": "red",
    "CRITICAL": "red",
}


DEFAULT_COLOR = "silver"


class ColoredFormatter(logging.Formatter):
    """
    A formatter that allows colors to be placed in the format string.

    Intended to help in creating more readable logging output.
    """

    def __init__(self, fmt=None, datefmt=None, style="%", log_colors=None):
        self.log_colors = log_colors or DEFAULT_COLORS
        super(ColoredFormatter, self).__init__(fmt=fmt, datefmt=datefmt, style=style)

    def format(self, record):
        message = logging.Formatter.format(self, record)

        config = self.log_colors.get(record.levelname, DEFAULT_COLOR)
        if isinstance(config, str):
            config = {
                "color": config,
                "background": None,
                "underline": False,
                "bold": False,
            }

        elif not isinstance(config, dict):
            raise ValueError("color setting must be a string or dict")

        message = cs(message, config.get("color", DEFAULT_COLOR), config.get("background", None))
        if config.get("underline", False):
            message = message.underline()
        if config.get("bold", False):
            message = message.bold()

        return str(message)



# Copyright [2018-2020] Peter Krenesky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
from importlib import import_module

from ixian.config import CONFIG
from ixian.exceptions import ModuleLoadError, InvalidClassPath
from ixian.task import Task, VirtualTarget


logger = logging.getLogger(__name__)
CLASS_PATH_PATTERN = re.compile(r"(?P<module_path>.*)\.(?P<classname>.+)")
MODULES = {}


def load_module(module_path):
    """
    Load module by path.

    This will load a module's config.md into CONFIG and tasks into cli.

    :param module_path: dot path to module package
    :return:
    """

    # load module and it's options
    module = import_module(module_path)
    if not hasattr(module, "OPTIONS"):
        raise ModuleLoadError(f"{module_path}: missing OPTIONS")
    MODULE_OPTIONS = getattr(module, "OPTIONS")

    # verify required options
    if "name" not in MODULE_OPTIONS and not hasattr(MODULE_OPTIONS, "name"):
        raise ModuleLoadError(f"{module_path}: OPTIONS is missing 'name'")

    # load config
    config_class_path = MODULE_OPTIONS.get("config", False)
    if config_class_path:
        match = CLASS_PATH_PATTERN.match(config_class_path)
        if not match:
            raise InvalidClassPath("Config classpath invalid: %s" % config_class_path)
        config_module_path, config_classname = match.groups()

        config_module = import_module(config_module_path)
        config_class = getattr(config_module, config_classname)
        CONFIG.add(MODULE_OPTIONS["name"].upper(), config_class())

    # load tasks
    tasks_module_path = MODULE_OPTIONS.get("tasks", False)
    if tasks_module_path:
        load_tasks(tasks_module_path)

    # add config to global so downstream utils/modules may use it
    MODULES[MODULE_OPTIONS["name"]] = MODULE_OPTIONS

    logger.debug("Loaded Module: {}".format(MODULE_OPTIONS["name"]))


def load_tasks(tasks_module_path):
    """Load a module with tasks in it.

    Args:
        tasks_module_path: module path as dot notation string.
    """
    tasks_module = import_module(tasks_module_path)
    for module_attribute in tasks_module.__dict__.values():
        if (
            isinstance(module_attribute, type)
            and issubclass(module_attribute, Task)
            and module_attribute not in (Task, VirtualTarget)
        ):
            try:
                module_attribute()
            except:
                logger.error("Error loading task: %s" % module_attribute)
                raise



# Copyright [2018-2020] Peter Krenesky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import typing

from ixian.check.checker import hash_object
from ixian.config import CONFIG
from ixian.exceptions import AlreadyComplete
from ixian.utils.color_codes import BOLD_WHITE, ENDC, GRAY, OK_GREEN


logger = logging.getLogger(__name__)
TASKS = {}


class TaskRunner(object):
    """
    A task is a wrapper around functions that adds in various functionality
    such as dependencies and check functions.

    func - function to run
    category - category to add this task to
    check - list of checkers
    clean - clean before building
    config - list of relevant settings to display in task help
    name - name of this task
    short_description - short description for task
    description - long description for task
    parent - parent task (this task will be run first if it is run)
    children - tasks this task depends on (They will be run first)
    """

    checkers = None

    def __init__(
        self,
        task=None,
        func=None,
        category=None,
        check=None,
        clean=None,
        config=None,
        depends=None,
        name=None,
        parent=None,
        short_description=None,
        description=None,
    ):
        self.task = task
        self.func = func
        self._depends = depends or []
        self.category = category.upper() if category else None
        self.clean = clean
        self.short_description = short_description or ""
        self.config = config

        # determine task name
        if name is not None:
            self.name = name

        # determine description
        self.description = description

        # Add task to global registry. Merge virtual target's dependencies if they exist.
        if self.name in TASKS:
            task_instance = TASKS[self.name]
            # The task is virtual if there is no func, replace it.
            if task_instance.func is None:
                self.add_dependency(*task_instance._depends)
                task_instance = self
            else:
                logger.warning("Duplicate task definition: {}".format(self.name))
        else:
            task_instance = self
        TASKS[self.name] = task_instance

        # add task to VirtualTargets if a parent is specified
        if parent:
            for parent in parent if isinstance(parent, list) else [parent]:
                self.add_to_parent(parent)

        # Setup checkers, clean method
        if check:
            if isinstance(check, (list, tuple)):
                self.checkers = check
            else:
                self.checkers = [check]

    def __str__(self):
        return f"<{type(self).__name__}@{id(self)} func={self.name}>"

    def __unicode__(self):
        return f"<{type(self).__name__}@{id(self)} func={self.name}>"

    def __repr__(self):
        return f"<{type(self).__name__}@{id(self)} func={self.name}>"

    @property
    def in_context(self):
        if not self.task:
            return False

        return self.task.contexts is True or CONFIG.RUN_CONTEXT in self.task.contexts

    def add_to_parent(self, name: str):
        """Add a task to as a dependency of a another task.

        This is a grouping method that allows modules to inject
        dependencies into common targets.

        If the target task is not defined a no-op task will be created to wrap
        the added tasks.

        :param name: name of parent task to add task to
        :return: parent task
        """
        try:
            parent = TASKS[name]
        except KeyError:
            # VirtualTarget wasn't defined explicitly, or task hasn't been loaded yet.
            # create a TaskRunner for the target. If an explicit task is loaded after
            # it will replace this and assume the children that were already added.
            parent = TaskRunner(name=name)
            TASKS[name] = parent
        parent.add_dependency(self)
        return parent

    def __call__(self, *args, **kwargs):
        return self.execute(args, **kwargs)

    def execute(self, args, **kwargs):
        """Execute this task.

        Executes this task including any dependencies that fail their checks.
        If a dependency fails it's check then this task will execute even if
        it's own checks pass.

        Tasks and dependencies may be forced by passing `force=True` or
        `force-all=True` as kwargs.

        Tasks and dependency clean methods may be run by passing `clean=True`
        or `clean-all=False` as kwargs. Clean implies `force=True`.

        :param args: args to pass through to the task
        :param kwargs: options for task execution
        :return: return value from task function
        """
        clean_root = kwargs.get("clean", False)
        clean_all = kwargs.pop("clean_all", False)
        force_root = kwargs.pop("force", False)
        force_all = kwargs.pop("force_all", False)

        if clean_root:
            force_root = True
        if clean_all:
            clean_root = True
            force_all = True
        if force_all:
            force_root = True

        # save force to task instance so it may be referenced downstream
        # TODO: this should be passing in `force`
        self.force = True

        args_as_str = CONFIG.format(" ".join([str(arg) for arg in args]))
        logger.debug(f"[exec] {self.name}({args_as_str}) force={force_root} clean={clean_root}")

        def execute_node(node, clean, force, args=None):
            runner = TASKS[node["name"]]

            if runner and runner.clean and clean:
                logger.debug(f"Cleaning Task: {runner.clean}")
                runner.clean()

            complete_dependencies = 0
            for dependency in node["dependencies"]:
                try:
                    execute_node(dependency, clean_all, force_all)
                except AlreadyComplete:
                    complete_dependencies += 1
            dependencies_complete = complete_dependencies == len(node["dependencies"])

            # Execute function if there is one. Targets may not have a function. If any dependency
            # was run, then this task must run too.
            if runner and runner.func:
                passes, checkers = runner.check(force)
                if dependencies_complete and passes:
                    logger.debug(f"[skip] {node['name']}, already complete.")
                    raise AlreadyComplete()

                else:
                    # set tasks force attribute so it's setup the same as if it were run directly.
                    runner.task.__task__.force = force
                    return_value = runner.func(*args or [])
                    # save checker only after function has completed successfully. Save should be
                    # called even if force=True
                    if checkers:
                        for checker in checkers:
                            checker.save()
                    logger.debug(f"[fini] {runner.name}")
                    return return_value

        return execute_node(self.tree(flatten=False), clean_root, force_root, args)

    def check(self, force: bool = False) -> (bool, list):
        """Return True if the task is complete based on configured checks.

        If the task does not have a checker this method always returns `False`.

        :param force: override the check and return True if True.
        :return:
        """
        checkers = [checker.clone() for checker in self.checkers] if self.checkers else None
        passes = False
        if self.checkers:
            if force:
                passes = False
            else:
                checks = [checker.check() for checker in checkers]
                passes = all(checks)
        return passes, checkers

    def state(self, shallow: bool = True) -> typing.Optional[dict]:
        """
        Calculates a dict of state generated from the tasks checkers.

        :param shallow: only return hash for dependencies
        :return: dict of state returned from checkers
        """
        if self.checkers is None and self.depends is None:
            return None

        checkers = (
            [checker.clone() for checker in self.checkers if checker.contribute_to_task_state]
            if self.checkers
            else None
        )

        depends = {}
        for dependency in self.depends:
            name = dependency.name
            if shallow:
                depends[name] = dependency.hash()
            else:
                depends[name] = dependency.state()

        return {
            "depends": depends,
            "checks": [
                {
                    "class": f"{type(checker).__module__}.{type(checker).__name__}",
                    "state": checker.state(),
                }
                for checker in checkers
            ],
        }

    def hash(self):
        return hash_object(self.state(shallow=True))

    def add_dependency(self, *tasks):
        self._depends.extend(tasks)

    @property
    def depends(self) -> list:
        return [
            dependency if isinstance(dependency, TaskRunner) else TASKS[CONFIG.format(dependency)]
            for dependency in self._depends
        ]

    def render_help(self, buffer) -> None:
        """render the "help" command

        Renders ixian internal help for the task. This help should explain
        how to use the task via ixian.

        Many tasks are proxies to other tools (e.g. npm, pipenv, etc). This
        help shouldn't try to replace that. Proxy tasks should indicate as such
        and include an example how to reach the tool's built-in help (--help)

        combines:
          - Name of task
          - Docstring as length description
          - task status tree
        """
        from ixian.config import CONFIG

        buffer.write(BOLD_WHITE)
        buffer.write("NAME\n")
        buffer.write(ENDC)
        buffer.write(f"    {self.name} -- {self.short_description}\n")
        buffer.write(BOLD_WHITE)
        buffer.write("\nDESCRIPTION\n")
        buffer.write(ENDC)
        if self.description:
            buffer.write(CONFIG.format(self.description))

        if self.config:
            buffer.write(BOLD_WHITE)
            buffer.write("\nCONFIGURATION\n")
            buffer.write(ENDC)
            padding = max(len(config) for config in self.config) - 1
            for config in self.config:
                buffer.write(
                    "    - {key}  {value}\n".format(
                        key="{key}:".format(key=config[1:-1]).ljust(padding),
                        value=CONFIG.format(config),
                    )
                )

        buffer.write(BOLD_WHITE)
        buffer.write("\n\nSTATUS\n")
        buffer.write(ENDC)
        self.render_status(buffer)

    def render_status(self, buffer) -> None:
        """render task status.

        Display the dependency tree for the task.

        Formatting/Readability optimizations:
         - Tree trimming: Redundant nodes are trimmed from the status tree.
            If A and B both depend on C then C will only be shown once.
        """

        def render_task(node, indent=0):
            # render task status
            if node["name"] is not None:
                passes = node["passes"]
                if passes:
                    icon = OK_GREEN + "✔" + ENDC
                else:
                    icon = GRAY + "○" + ENDC
                if indent:
                    spacer = "  " * indent
                else:
                    spacer = ""

                task_line = f'{spacer}{icon} {node["name"]}\n'
                buffer.write(task_line)
                indent += 2

            for dependency in node["dependencies"]:
                render_task(dependency, indent=indent)

        render_task(self.status(), indent=0)

    def tree(self, dedupe: bool = True, flatten: bool = True) -> dict:
        """
        Return tree of tasks, with this task as the root.

        :param dedupe: remove duplicates from tree
        :param flatten: flatten single item dependcy lists into the parent
        :return:
        """
        tree = self._build_tree(set([]) if dedupe else None)
        if flatten:
            tree = flatten_tree(tree)
        return tree

    def _build_tree(self, seen=None):
        """
        Internal method for recursively building task tree
        :param seen: should be a Set if deduping.
        :return: node in tree
        """
        dependencies = []
        for dependency in self.depends:
            if seen is not None:
                if dependency in seen:
                    continue
                seen.add(dependency)
            dependencies.append(dependency._build_tree(seen))

        return {"name": self.name, "dependencies": dependencies}

    def status(self, dedupe: bool = True, flatten: bool = True) -> dict:
        """
        Return the task tree augmented with status information
        """

        def update(node):
            for dependency in node["dependencies"]:
                update(dependency)

            if node["name"] is not None:
                # Run self.check even if children fail their checks. That way the
                # checkers (and state) are available.
                children_passes = all(
                    (dependency["passes"] for dependency in node["dependencies"])
                )
                runner = TASKS[node["name"]]
                passes, checkers = runner.check()
                node["checkers"] = checkers

                # node fails if any children have failed
                node["passes"] = passes and children_passes

            return node

        return update(self.tree(dedupe, flatten))


def flatten_tree(tree: dict, full: bool = False) -> dict:
    """
    Flatten an execution tree to make it easier to read.

    Task trees are often a single node nested several levels deep. These trees may be collapsed
    into a list. The execution order is the same, but it's easier for a human to read.

    Before:
        - foo
          - bar
            - xoo

    After:
        - xoo
        - bar
        - foo


    Before:
        - foo
          - xar
          - bar
            - xoo

    After:
        - foo
          - xar
          - xoo
          - bar

    :param tree: Tree to flatten
    :param full: Flatten tree into single list
    :return: flattened task list
    """

    def flatten_node(node: dict) -> list:
        """
        Flatten a single node. Always return a list for consistency, even when returning a single
        node.

        :param node:
        :param parent: parent task list to collapse into
        :return: flattened node
        """
        node = node.copy()
        num_dependencies = len(node["dependencies"])
        if num_dependencies == 0:
            # no dependencies: nothing to flatten, return as-is
            return [node]

        elif full or num_dependencies == 1:
            # flatten dependencies: flatten into single list that includes parent & child
            flattened = []
            for dependency in node["dependencies"]:
                flattened_child = flatten_node(dependency)
                flattened.extend(flattened_child)

                # clear dependencies, since they are now siblings
                # this node is added last since it runs after dependencies
                node["dependencies"] = []
                flattened.append(node)

            return flattened

        else:
            # multiple dependencies: do not flatten into parent.
            #
            # Any dependencies that are flattened need to be merged with other dependencies.
            # Dependency nodes should either be a single node, or a list of nodes
            dependencies = []
            for dependency in node["dependencies"]:
                flattened = flatten_node(dependency)
                dependencies.extend(flattened)

            node["dependencies"] = dependencies
            return [node]

    root = flatten_node(tree)
    if len(root) > 1:
        # if root's dependencies were flattened into it, then the returned list will have all of
        # those dependencies. Create a new root node to contain them all. This keeps the structure
        # consistent-ish for downstream consumers. They still have to special case this node, but
        # it should be a little simpler since all nodes are of a similar shape
        return {"name": None, "dependencies": root}
    else:
        # a single node, unpack it and return as root.
        return root[0]


class Task(object):
    """
    Super class for defining ixian tasks.

    Task subclasses should define an execute method.
    """

    __task__ = None
    contexts = ["cli"]

    @property
    def __func__(self):
        if not hasattr(self, "execute"):
            raise NotImplementedError("Task classes must implement execute method")

        # wrap execute method to curry `self`
        def execute(*args, **kwargs):
            return self.execute(*args, **kwargs)

        return execute

    def __new__(cls, *args, **kwargs):
        instance = super(Task, cls).__new__(cls, *args, **kwargs)

        # TODO: fix needed, for broken tests this causes
        # if instance.name not in TASKS:
        if cls.__task__ is None not in TASKS:
            cls.__task__ = TaskRunner(
                task=instance,
                func=instance.__func__,
                name=instance.name,
                category=getattr(instance, "category", None),
                depends=getattr(instance, "depends", None),
                check=getattr(instance, "check", None),
                clean=getattr(instance, "clean", None),
                config=getattr(instance, "config", None),
                parent=getattr(instance, "parent", None),
                short_description=getattr(instance, "short_description", None),
                description=cls.__doc__,
            )
        else:
            # In practice task classes should never need to be instantiated more than once.
            # Unloading tasks isn't supported at this time, but tests may do that. When that
            # happens subsequent tests may see this fail. This msg helps show that happened.
            # hopefully this is fixed in a better way when task loading/tree is refactored.
            logger.warning(f"Task {instance.name} instantiated but an instance already exists")

        return instance

    def __call__(self, *args, **kwargs):
        type(self).__task__(*args, **kwargs)


class VirtualTarget(Task):
    """
    A virtual target is a placeholder task that is used for targets that
    don't have a concrete task registered. VirtualTargets may be executed the
    same as tasks. When run, they execute dependencies that were registered
    with them.

    VirtualTargets allow the target to be given a description, dependencies,
    and other options. VirtualTargets allow tasks grouping without tight
    coupling to a specific target.

    Tasks and other VirtualTargets register with another VirtualTarget by
    specifying the targets as it's parent.  I.e. `parent='my_target'`.

    If multiple modules implement VirtualTargets with the same name, then they
    will be merged. This allows modules to define the same groupings.

    For example, javascript and python modules might both define a `test`
    target to encapsulate all tests. Build pipeline tools can be built to
    expect the generic `test` target regardless of whether a project use
    python, javascript, or any other combination of languages.

    If a concrete Task with the same name as VirtualTarget is registered, the
    Task will replace the VirtualTarget. Tasks that contribute to the virtual
    target act as dependencies, they'll run before any concrete task.
    """

    @property
    def __func__(self):
        return None



# Copyright [2018-2020] Peter Krenesky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import inspect
import json

import os

from ixian.config import CONFIG
from ixian.utils.filesystem import mkdir


def hash_object(obj):
    """deterministically hash a dict."""
    hash = hashlib.sha256()
    hash.update(json.dumps(obj, sort_keys=True).encode("utf-8"))
    return hash.hexdigest()


class Checker(object):
    """
    Checkers determine whether a task should run or be skipped. Tasks may have
    zero or more checkers that run before it is executed. The task will skip
    when the checker indicates the task is already compete.

    Each checker should implement the `data` method which returns an object
    that represents the state being checked. The state object is serialized and
    saved after successful runs. The current state is compared against this to
    determine whether the task should run.
    """

    contribute_to_task_state = True

    def check(self):
        return self.state() == self.saved_state()

    # TODO optimize to avoid repeated state calculation
    def state(self):
        """
        State being considered as part of this check. State is expected to be
        a dict of properties and their values. State must be json serializable
        and sorted.

        State values may be any type of json serializable object. For example,
        a file_hasher would use a hash of the file.

        The set of keys in a state object should be idempotent regardless of
        state.

        :return: state dict
        """
        raise NotImplementedError

    def hash(self):
        return hash_object(self.state())

    def saved_state(self):
        """
        Retrieved saved state, if any.  If task has not been run successfully
        it's state will be None.
        :return: state object
        """
        file_path = self.file_path()
        if os.path.exists(file_path):
            with open(file_path) as file:
                return json.loads(file.read())
        else:
            return None

    def file_path(self):
        """Path where state file can be found.

        :return: path
        """
        return CONFIG.format("{BUILDER}/checks/{file_name}", file_name=self.filename())

    def filename(self):
        """
        Generate a filename that is a deterministic representation of the
        dependencies being checked.
        """
        raise NotImplementedError

    def save(self):
        mkdir(CONFIG.format("{BUILDER}/checks/"))
        state = self.state()
        with open(self.file_path(), "w") as file:
            file.write(json.dumps(state))

    def clone(self):
        raise NotImplementedError


class MultiValueChecker(Checker):
    """Checker that checks multiple keys."""

    def __init__(self, *keys):
        assert len(keys) != 0, "At least one key must be given"
        self._keys = keys

    @property
    def keys(self):
        """list of keys saved in this instance"""
        return [CONFIG.format(key) for key in self._keys]

    def filename(self):
        """"Generate file path using keys of the data dict."""
        return hash_object(self.keys)

    def clone(self):
        return type(self)(*self._keys)



# Copyright [2018-2020] Peter Krenesky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

BOLD_WHITE = "\033[1m"
RED = "\033[91m"
OK_GREEN = "\033[92m"
ENDC = "\033[0m"
YELLOW = "\033[93m"
GRAY = "\033[90m"


COLOR_REFERENCE = {
    "BOLD_WHITE": BOLD_WHITE,
    "RED": RED,
    "OK_GREEN": OK_GREEN,
    "YELLOW": YELLOW,
    "GRAY": GRAY,
}


def format(txt, color):
    return f"{color}{txt}{ENDC}"


def red(txt):
    return format(txt, RED)


def green(txt):
    return format(txt, OK_GREEN)


def yellow(txt):
    return format(txt, YELLOW)


def gray(txt):
    return format(txt, GRAY)


def bold_white(txt):
    return format(txt, BOLD_WHITE)



# Copyright [2018-2020] Peter Krenesky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import os
from glob import glob
from itertools import chain
from stat import ST_MODE

from ixian.check.checker import MultiValueChecker, hash_object


def get_flags(path):
    """Return permissions and flags in a format suitable for hashing"""
    return str(os.stat(path)[ST_MODE])


def hash_file(path):
    """
    Hash a single file including it's contents, permissions, and flags.

    :param path: path to file or directory
    :return: sha256 hash of path contents
    """
    hash = hashlib.sha256()
    hash.update(get_flags(path).encode("utf-8"))
    with open(path, "rb", buffering=0) as f:
        for b in iter(lambda: f.read(128 * 1024), b""):
            hash.update(b)
    return hash.hexdigest()


def hash_dir(path):
    """ Hash the contents, permissions and flags of a directory. Subdirectories
    are recursed into and hashed.

    Filename's of contents are included in the hash. Renaming a file results in
    a different hash for the directory containing it.

    Only the relative path of contents is considered. Directories with
    identical contents and renamed directories will return the same hash.

    :param path: path of directory
    :return: sha256 hash of directory contents
    """
    content_hashes = {"___FLAGS___": get_flags(path)}
    contents = sorted(os.listdir(path))
    for child in contents:
        child_path = os.path.join(path, child)
        if os.path.isdir(child_path):
            content_hashes[child] = hash_dir(child_path)
        else:
            content_hashes[child] = hash_file(child_path)
    return hash_object(content_hashes)


def hash_path(path):
    """Hash file or directory

    :param path: path to hash
    :return: sha256 hash
    """
    if os.path.isdir(path):
        return hash_dir(path)
    else:
        return hash_file(path)


def hash_paths(*paths):
    """ hash directories and files
    :param paths: list of directories or file paths
    :return: dict mapping path to hash
    """
    return {path: hash_path(path) for path in paths}


class FileHash(MultiValueChecker):
    """Checker that hashes files and directories.

    Keys are paths that may include unix style wildcards and config variables.

    :Example:

    Checker(
       '/path/to/my/file',
       '/path/to/my/other/file',
       '/path/to/my/directory',
       '/wildcard/*'
    )

    Directory contents are recursively hashed.
    """

    @property
    def keys(self):
        """Returns paths, expanding any wildcards into matches."""
        # TODO don't expand wildcards here. Expanded set of files are not static
        # files may be removed or created. This would cause a different set of
        # keys and a different filename. The filename should just be raw keys
        # and expand them below.
        _keys = super(FileHash, self).keys
        expanded = (glob(pattern) for pattern in _keys)
        return list(chain(*expanded))

    def state(self):
        return hash_paths(*self.keys)



# Copyright [2018-2020] Peter Krenesky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from enum import Enum
from logging.config import dictConfig

import argcomplete
import argparse
import importlib
import importlib.machinery
import io
import types
import sys
from collections import defaultdict

from ixian.config import CONFIG
from ixian.module import load_module
from ixian.utils import filesystem as file_utils
from ixian.exceptions import AlreadyComplete, ExecuteFailed
from ixian.task import TASKS, TaskRunner
from ixian.utils.color_codes import RED, ENDC
from ixian.utils.decorators import classproperty


logger = logging.getLogger(__name__)


class ExitCodes(Enum):
    """Codes that may be returned by cli"""

    SUCCESS = 0  # Success
    ERROR_COMPLETE = -1  # task is already complete
    ERROR_UNKNOWN_TASK = -2  # task isn't registered
    ERROR_NO_INIT = -3  # ixian.py does not contain an init flag
    ERROR_NO_IXIAN_PY = -4  # ixian.py does not exist
    ERROR_TASK = -5  # task did not complete

    @classproperty
    def errors(cls):
        return [e for e in list(cls) if e.is_error]

    @classproperty
    def init_errors(cls):
        """Errors that init can raise"""
        return [
            cls.ERROR_NO_IXIAN_PY,
            cls.ERROR_NO_INIT,
        ]

    @classproperty
    def run_errors(cls):
        """Errors that run can raise"""
        return [
            cls.ERROR_UNKNOWN_TASK,
            cls.ERROR_COMPLETE,
            cls.ERROR_TASK,
        ]

    @property
    def is_success(self):
        return self == self.SUCCESS

    @property
    def is_error(self):
        return self != self.SUCCESS


def ixian_path() -> str:
    """Return path to ixian.py"""
    env_value = os.getenv("IXIAN_CONFIG", None)
    if env_value:
        return env_value
    else:
        return f"{file_utils.pwd()}/ixian.py"


def import_ixian():
    """Imports a ixian module and returns it."""
    path = ixian_path()

    loader = importlib.machinery.SourceFileLoader("ixian", f"{path}")
    ixian_module = types.ModuleType(loader.name)
    loader.exec_module(ixian_module)
    return ixian_module


def init() -> ExitCodes:
    """init ixian

    Finds the ixian config module for the projects and initializes itself
    using the `setup` function found inside ixian.py.

    :return:
    """
    init_logging()

    try:
        ixian_module = import_ixian()
    except FileNotFoundError as e:
        logger.error(str(e))
        return ExitCodes.ERROR_NO_IXIAN_PY

    try:
        module_init = ixian_module.init
    except AttributeError:
        logger.error("init() was not found within ixian.py")
        return ExitCodes.ERROR_NO_INIT

    # init module and return all globals.
    logger.debug("Ixian v0.0.1")
    load_module("ixian.modules.core")
    module_init()

    return ExitCodes.SUCCESS


def build_epilog() -> str:
    """Build help epilog text"""
    output = io.StringIO()
    if TASKS:
        categories = defaultdict(list)
        for task in TASKS.values():
            # only list tasks for the current context
            if task.in_context:
                categories[task.category].append(task)
        padding = max(len(task.name) for task in TASKS.values())
        output.write(
            """Type 'ix help <subcommand>' for help on a specific """ """subcommand.\n\n"""
        )
        output.write("""Available subcommands:\n\n""")

        for name, tasks in categories.items():
            category = name.capitalize() if name else "Misc"
            output.write(f"{RED}[ {category} ]{ENDC}\n")
            for task in sorted(tasks, key=lambda t: t.name.upper()):
                task_name = task.name.ljust(padding, " ")
                output.write(f"  {task_name}    {task.short_description}\n")
            output.write("\n")

    return output.getvalue()


def init_logging() -> None:
    """Initialize logging system."""
    args = parse_args()

    dictConfig(CONFIG.LOGGING_CONFIG)
    root = logging.getLogger()
    root.setLevel(args["log"])


def get_parser() -> argparse.ArgumentParser:
    """Return an instance of the base argument parser.

    The base parser has the internal flags but not tasks.
    :return: ArgumentParser that can parse args.
    """
    parser = argparse.ArgumentParser(
        prog="ixian",
        add_help=False,
        description="Run a ixian task.",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=build_epilog(),
    )
    # TODO: try to fix formatting for choices
    parser.add_argument("--help", help="show this help message and exit", action="store_true")
    parser.add_argument(
        "--log", type=str, help="Log level (DEBUG|INFO|WARN|ERROR|NONE)", default="DEBUG",
    )
    parser.add_argument("--force", help="force task execution", action="store_true")
    parser.add_argument(
        "--force-all", help="force execution including task dependencies", action="store_true",
    )
    parser.add_argument("--clean", help="clean before running task", action="store_true")
    parser.add_argument(
        "--clean-all", help="clean all dependencies before running task", action="store_true",
    )
    parser.add_argument("remainder", nargs=argparse.REMAINDER, help="arguments for task.")
    return parser


DEFAULT_ARGS = {
    "clean": False,
    "clean_all": False,
    "force": False,
    "force_all": False,
    "log": "DEBUG",
    "task": "help",
    "task_args": None,
    "help": False,
}


def parse_args(args: str = None) -> dict:
    """Parse args from command line input"""
    parser = get_parser()
    compiled_args = DEFAULT_ARGS.copy()
    parsed_args, _ = parser.parse_known_args(args)
    compiled_args.update(parsed_args.__dict__)
    remainder = compiled_args.pop("remainder")

    if remainder:
        compiled_args["task"] = remainder[0]
        compiled_args["task_args"] = remainder[1:]
    else:
        compiled_args["task"] = "help"
        compiled_args["task_args"] = []

    compiled_args["log"] = compiled_args["log"]

    # if --help flag is given then run the "help" task. "--help <task>" and "help <task>" should
    # be treated the same.
    if parsed_args.help:
        # When --help is used the first arg is the task to show help for. The parser always saves
        # the first arg as "task". --help should always run "help" as the task. Move the task
        # into `task_args` so it is passed to the help task.
        if compiled_args["task"] and compiled_args["task"] != "help":
            compiled_args["task_args"] = [compiled_args["task"]]
        compiled_args["task"] = "help"

    return compiled_args


"""
optional arguments:
  -h, --help   show this help message and exit
  --log LOG    Log level (DEBUG|INFO|WARN|ERROR|NONE)
  --force      force task execution
  --force-all  force execution including task dependencies
  --clean      clean before running task
  --clean-all  clean all dependencies before running task

Type ix help <subcommand>' for help on a specific subcommand.
"""


# TODO: logging isn't working in this task
def resolve_task(key: str) -> TaskRunner:
    """
    Resolve a task from the register by it's name

    :param key: name of task to resolve
    :return: TaskRunner instance
    """
    try:
        task = TASKS[key]
    except KeyError:
        logger.error('Unknown task "%s", run with --help for list of commands' % key)
    else:
        if task.in_context:
            return task
        else:
            logger.debug(
                f"RUN_CONTEXT ({CONFIG.RUN_CONTEXT}) does not match task config ({task.task.contexts})"
            )


def load_environment():
    """
    Loads environment variables and updates CONFIG with them.
    """
    for key, value in os.environ.items():
        if key.startswith(CONFIG.ENV_PREFIX):
            parts = key[len(CONFIG.ENV_PREFIX) :].split("__")
            current = CONFIG
            for part in parts[:-1]:
                current = getattr(current, part)
            setattr(current, parts[-1], value)


def run() -> ExitCodes:
    """
    Run ixian task

    :return: 0 if successful or an error code
    """

    # parse args
    args = parse_args()
    task_name = args.pop("task")
    task_args = args.pop("task_args")
    formatted_task_args = [CONFIG.format(arg) for arg in task_args]

    # TODO: load environment here for now. Eventually move this into init() so env is available
    #       as modules load. Dynamic loading like that is more complex
    load_environment()

    task = resolve_task(task_name)
    if not task:
        return ExitCodes.ERROR_UNKNOWN_TASK

    try:
        task.execute(formatted_task_args, **args)
    except AlreadyComplete:
        logger.warning("Already complete. Override with --force or --force-all")
        return ExitCodes.ERROR_COMPLETE
    except ExecuteFailed as e:
        logger.error(str(e))
        return ExitCodes.ERROR_TASK

    return ExitCodes.SUCCESS


def exit_with_code(code):
    if isinstance(code, ExitCodes):
        code = code.value
    sys.exit(code)


def cli() -> None:
    """
    Main entry point into the command line interface.
    """
    init_code = init()
    if init_code.is_error:
        exit_with_code(init_code)

    # setup autocomplete
    parser = get_parser()
    argcomplete.autocomplete(parser)

    # run ixian
    run_code = run()
    if run_code is not None and run_code.is_error:
        exit_with_code(run_code)
    else:
        exit_with_code(ExitCodes.SUCCESS)



# Copyright [2018-2020] Peter Krenesky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import List, Dict, Union


# TODO: type hints for arg_groups and return
def merge_parser_args(
    parser: argparse.ArgumentParser, *arg_groups: Union[List[str]]
) -> Dict[str, str]:
    """
    Parse multiple arg lists with the same parser and merged them.

    :param parser:
    :param arg_groups:
    :return:
    """
    merged_options = {}
    for args in arg_groups:
        options, _ = parser.parse_known_args(args)
        merged_options.update({k: v for k, v in vars(options).items() if v is not None})
    return merged_options


def argunparse(options: dict, parser: argparse.ArgumentParser) -> List[str]:
    """
    Convert a dict of flags back into a list of args.
    """
    args = []
    for argument in parser.arguments:
        single_dash_name = next((arg for arg in argument["args"] if arg.startswith("-")), None)
        double_dash_name = next((arg for arg in argument["args"] if arg.startswith("--")), None)
        has_double_dash = bool(double_dash_name)
        flag_name = double_dash_name if has_double_dash else single_dash_name

        if "dest" in argument:
            value = options[argument["dest"]]
        else:
            for flag in argument["args"]:
                flag_stripped = flag.lstrip("-").replace("-", "_")
                if flag_stripped in options:
                    value = options[flag_stripped]
                    break
            else:
                value = None

        if value:
            # flags without values
            # TODO: doesn't handle flags that set a non-bool const value
            if isinstance(value, bool):
                as_args = [flag_name]

            # flags with values
            else:
                if has_double_dash and not argument.get("nargs", None):
                    arg_template = "{flag}={value}".format(flag=flag_name, value="{value}")
                else:
                    arg_template = "{flag} {value}".format(flag=flag_name, value="{value}")

                if not isinstance(value, list):
                    as_args = [arg_template.format(value=value)]
                else:
                    # for now assume separate flags per value e.g. "--foo bar --foo xoo"
                    # also need to support single flag with multiple values "--foo bar xoo"
                    as_args = [arg_template.format(value=v) for v in value]
            args.extend(as_args)

    return args



# Copyright [2018-2020] Peter Krenesky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ClassPropertyDescriptor(object):
    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)  # pragma: no cover
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:  # pragma: no cover
            raise AttributeError("can't set attribute")  # pragma: no cover
        type_ = type(obj)  # pragma: no cover
        return self.fset.__get__(obj, type_)(value)  # pragma: no cover

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def classproperty(func):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)


class cached_property:
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance.

    Optional ``name`` argument allows you to make cached properties of other
    methods. (e.g.  url = cached_property(get_absolute_url, name='url') )
    """

    def __init__(self, func, name=None):
        self.func = func
        self.__doc__ = getattr(func, "__doc__")
        self.name = name or func.__name__

    def __get__(self, instance, cls=None):
        """
        Call the function and put the return value in instance.__dict__ so that
        subsequent attribute access on the instance returns the cached value
        instead of calling cached_property.__get__().
        """
        if instance is None:
            return self  # pragma: no cover
        res = instance.__dict__[self.name] = self.func(instance)
        return res



# Copyright [2018-2020] Peter Krenesky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import re

from ixian.utils.filesystem import pwd
from ixian.utils.decorators import classproperty


logger = logging.getLogger(__name__)


class MissingConfiguration(AssertionError):
    def __init__(self, key, parent=None):
        """
        Error thrown when formatting a string but a config value is missing.
        :param parent: Key being rendered
        :param key: Key that is missing
        """
        if parent:
            msg = f"Missing config while rendering {parent}: {key}"
        else:
            msg = f"Missing config: {key}"
        super(MissingConfiguration, self).__init__(msg)


CONFIG_VARIABLE_PATTERN = re.compile(r"{(?P<var>[a-zA-Z0-9_.]+)}")


class Config(object):
    root = None
    reserved = {"add", "format", "reserved", "children", "root", "__dict__"}

    def __getattribute__(self, key):
        """
        Overloaded to implement recursive lazy evaluation of properties.
        :param key: key to get
        :return: value if exists.
        """
        value = super(Config, self).__getattribute__(key)

        if key == "reserved" or key in self.reserved:
            return value
        else:
            return self.format(value, key)

    def add(self, key, child_config):
        """
        Add a child config
        :param key: Key to add child config as. e.g. "PYTHON"
        :param child_config: config to add.
        """
        self.__dict__[key] = child_config
        child_config.root = self

    def format(self, value, key=None, **kwargs):
        """
        Format strings using CONFIG object.

        This method uses python's built-in `str.format()` method. All root
        properties in CONFIG are passed in as kwargs. The properties lazy
        evaluate and recursively expand.

        Example:
            "{HOST}:{PORT}" may be formatted "0.0.0.0:8000"

        :param value: string to format
        :param key: TODO: is this needed?
        :param **kwargs: additional kwargs to format
        """
        if not isinstance(value, str):
            return value

        # always format strings using the root so the full path is available
        if self.root:
            return self.root.format(value, key, **kwargs)

        variables = CONFIG_VARIABLE_PATTERN.findall(value)
        expanded = {}
        for variable in variables:
            if variable not in kwargs:
                try:
                    root_key = variable.split(".")[0]
                    root = self.root if self.root else self

                    expanded[root_key] = self.format(getattr(root, root_key), variable, **kwargs)
                except AttributeError:
                    raise MissingConfiguration(variable, key)

        expanded.update(**kwargs)
        return value.format(**expanded)

    def resolve(self, path):
        value = self
        for key in path.split("."):
            value = getattr(value, key)
        return value

    # =========================================================================
    #  Base config
    # =========================================================================
    @classproperty
    def IXIAN(cls):
        """Directory where ixian is installed"""
        import ixian

        return os.path.dirname(os.path.realpath(ixian.__file__))

    @classproperty
    def PWD(cls):
        """Directory where ixian was run from"""
        return pwd()

    PROJECT_NAME = None

    # ENV - build environment PRODUCTION or DEV
    ENV = "DEV"

    # Prefix for environment variables to autoload
    ENV_PREFIX = "JT_"

    # RUN_CONTEXT - run context, by default this is the cli
    RUN_CONTEXT = "cli"

    # Local store for task runtime data.
    BUILDER_DIR = ".builder"
    BUILDER = "{PWD}/{BUILDER_DIR}"

    LOG_LEVEL = "DEBUG"
    FORMATTER = "console"
    LOGGING_CONFIG = {
        "version": 1,
        "formatters": {
            "detailed": {
                "class": "logging.Formatter",
                "format": "%(asctime)s %(levelname)-8s %(message)s",
            },
            "console": {
                "()": "ixian.logger.ColoredFormatter",
                "format": "%(asctime)s %(message)s",
                "log_colors": {
                    "DEBUG": "grey",
                    "INFO": "silver",
                    "WARNING": "yellow",
                    "ERROR": "red",
                    "CRITICAL": {"color": "red", "background": "white"},
                },
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": FORMATTER,
                "level": "DEBUG",
            }
        },
        "root": {"level": "DEBUG", "handlers": [FORMATTER]},
    }

    @classproperty
    def TASKS(self):
        return TasksConfig()


class TasksConfig:
    reserved = {"reversed", "root", "__dict__"}

    def __getattribute__(self, key):
        """
        Overloaded to implement recursive lazy evaluation of properties.
        :param key: key to get
        :return: value if exists.
        """
        from ixian.task import TASKS

        formatted_key = key.lower()
        if formatted_key in TASKS:
            task = TASKS[formatted_key]
            return TaskConfig(task)
        else:
            return super(TasksConfig, self).__getattribute__(key)


class TaskConfig:
    def __init__(self, task):
        self.task = task

    @property
    def STATE(self):
        return self.task.state()

    @property
    def HASH(self):
        return self.task.hash()


CONFIG = Config()



# Copyright [2018-2020] Peter Krenesky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import subprocess

from ixian.config import CONFIG
from ixian.exceptions import ExecuteFailed


logger = logging.getLogger(__name__)


def raise_for_status(code: int) -> None:
    """
    Raise `ExecuteFailed` if the code is an error code
    :param code: code to check
    """
    if code != 0:
        raise ExecuteFailed(f"Process returned a non-zero code: {code}")


def execute(command: str, silent: bool = False, env: dict = None) -> int:
    """
    Execute a shell command.

    ```
    execute("echo this is an example")
    ```

    Any config variables will be expanded.

    ```
    execute("echo this is the working directory: {PWD}")
    ```

    :param command: space separated command and args
    :param silent: do not echo command
    :return:
    """
    formatted_command = CONFIG.format(command)
    if not silent:
        logger.info(formatted_command)

    # Need to pass full environment, merge passed in env with process env.
    built_env = os.environ.copy()
    if env:
        built_env.update(env)

    args = [arg for arg in formatted_command.split(" ") if arg]
    return subprocess.call(args, env=built_env)


def get_dev_uid() -> int:
    """get dev uid of running process"""
    return int(subprocess.check_output(["id", "-u"])[:-1])


def get_dev_gid() -> int:
    """get dev gid of running process"""
    return int(subprocess.check_output(["id", "-g"])[:-1])



# Copyright [2018-2020] Peter Krenesky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io

from ixian.task import Task, VirtualTarget


class Lint(VirtualTarget):
    """Virtual target for linting project."""

    name = "lint"
    category = "testing"
    short_description = "Run all linting tasks."


class Test(VirtualTarget):
    """Virtual target for running all tests."""

    name = "test"
    category = "testing"
    short_description = "Run all testing tasks."


# =============================================================================
#  Teardown
# =============================================================================


class Clean(VirtualTarget):
    """Virtual target for cleaning the project."""

    name = "clean"
    category = "build"
    short_description = "Run all clean tasks."


class Help(Task):
    """
    Ixian internal help. Displays either internal help or a task's help.
    """

    name = "help"
    short_description = "This help message or help <task> for task help"
    contexts = True

    def execute(self, task_name=None):
        from ixian import runner

        if task_name:
            subtask = runner.resolve_task(task_name)
            buffer = io.StringIO()
            subtask.render_help(buffer)
            print(buffer.getvalue())
            buffer.close()
        else:
            parser = runner.get_parser()
            parser.print_help()
        return 0



# Copyright [2018-2020] Peter Krenesky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class AlreadyComplete(Exception):
    """
    Exception thrown when a Task executes but it's checks indicate complete.
    """


class ExecuteFailed(Exception):
    """
    Exception thrown by tasks when execute fails.
    """


class MockExit(BaseException):
    """Thrown by mock_exit to simulate exiting the process"""

    def __init__(self, code):
        self.code = code


class ModuleLoadError(Exception):
    """Error thrown when module can't be loaded"""

    pass


class InvalidClassPath(Exception):
    """Error thrown when a class path is not in the valid format"""

    pass
